
__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

import radical.utils as ru

from .base import ResourceManager


# ------------------------------------------------------------------------------
#
class LSF(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        ResourceManager.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        lsf_hostfile = os.environ.get('LSB_DJOB_HOSTFILE')
        if lsf_hostfile is None:
            msg = "$LSB_DJOB_HOSTFILE not set!"
            self._log.error(msg)
            raise RuntimeError(msg)

        lsb_mcpu_hosts = os.environ.get('LSB_MCPU_HOSTS')
        if lsb_mcpu_hosts is None:
            msg = "$LSB_MCPU_HOSTS not set!"
            self._log.error(msg)
            raise RuntimeError(msg)

        # parse LSF hostfile
        # format:
        # <hostnameX>
        # <hostnameX>
        # <hostnameY>
        # <hostnameY>
        #
        # There are in total "-n" entries (number of tasks)
        # and "-R" entries per host (tasks per host).
        # (That results in "-n" / "-R" unique hosts)
        #
        lsf_nodes = [line.strip() for line in open(lsf_hostfile)]
        self._log.info("Found LSB_DJOB_HOSTFILE %s. Expanded to: %s",
                       lsf_hostfile, lsf_nodes)
        lsf_node_list = list(set(lsf_nodes))

        # Grab the core (slot) count from the environment
        # Format: hostX N hostY N hostZ N
        lsf_cores_count_list = list(map(int, lsb_mcpu_hosts.split()[1::2]))
        lsf_core_counts      = list(set(lsf_cores_count_list))
        lsf_cores_per_node   = min(lsf_core_counts)
        lsf_gpus_per_node    = self._cfg.get('gpus_per_node', 0)  # FIXME GPU
        lsf_lfs_per_node     = {'path': ru.expand_env(
                                           self._cfg.get('lfs_path_per_node')),
                                'size':    self._cfg.get('lfs_size_per_node', 0)
                               }
        lsf_mem_per_node     = self._cfg.get('mem_per_node', 0)

        self._log.info("Found unique core counts: %s Using: %d",
                      lsf_core_counts, lsf_cores_per_node)

        # node names are unique, so can serve as node uids
        self.node_list        = [[node, node] for node in lsf_node_list]
        self.cores_per_node   = lsf_cores_per_node
        self.gpus_per_node    = lsf_gpus_per_node
        self.lfs_per_node     = lsf_lfs_per_node
        self.mem_per_node     = lsf_mem_per_node


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import copy
import time
import errno
import queue
import tempfile
import threading
import traceback

from orte_cffi import ffi, lib as orte_lib                                # noca

import radical.utils as ru

from ....  import pilot     as rp
from ...   import states    as rps
from ...   import constants as rpc
from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
#
def rec_makedir(target):

    # recursive makedir which ignores errors if dir already exists

    try:
        os.makedirs(target)

    except OSError as e:
        # ignore failure on existing directory
        if e.errno == errno.EEXIST and os.path.isdir(os.path.dirname(target)):
            pass
        else:
            raise


# ------------------------------------------------------------------------------
#
@ffi.def_extern()
def launch_cb(task, jdata, status, cbdata):
    return ffi.from_handle(cbdata).task_spawned_cb(task, status)


# ------------------------------------------------------------------------------
#
@ffi.def_extern()
def finish_cb(task, jdata, status, cbdata):
    return ffi.from_handle(cbdata).task_completed_cb(task, status)


# ------------------------------------------------------------------------------
#
class ORTE(AgentExecutingComponent):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__(self, cfg, session)

        self._watcher   = None
        self._terminate = threading.Event()


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)

        self._cancel_lock     = ru.RLock()
        self._tasks_to_cancel = list()
        self._watch_queue     = queue.Queue ()

        self._pid = self._cfg['pid']

        self.task_map = {}
        self.task_map_lock = ru.Lock()

        # we needs the LaunchMethod to construct commands.
        assert(self._cfg['task_launch_method'] ==
               self._cfg['mpi_launch_method' ] == "ORTE_LIB"), \
               "ORTE_LIB spawner only works with ORTE_LIB LaunchMethod."

        self._task_launcher = rp.agent.LaunchMethod.create(name="ORTE_LIB",
                                           cfg=self._cfg, session=self._session)
        self._orte_initialized = False
        self._task_environment   = self._populate_task_environment()

        self.gtod   = "%s/gtod" % self._pwd
        self.tmpdir = tempfile.gettempdir()


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        self._log.info('command_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_tasks':

            self._log.info("cancel_tasks command (%s)" % arg)
            with self._cancel_lock:
                self._tasks_to_cancel.extend(arg['uids'])

        return True


    # --------------------------------------------------------------------------
    #
    def _populate_task_environment(self):
        """Derive the environment for the task's from our own environment."""

        # Get the environment of the agent
        new_env = copy.deepcopy(os.environ)

        #
        # Mimic what virtualenv's "deactivate" would do
        #
        old_path = new_env.pop('_OLD_VIRTUAL_PATH', None)
        if old_path:
            new_env['PATH'] = old_path

        # TODO: verify this snippet from:
        # https://github.com/radical-cybertools/radical.pilot/pull/973/files
        # old_ppath = new_env.pop('_OLD_VIRTUAL_PYTHONPATH', None)
        # if old_ppath:
        #     new_env['PYTHONPATH'] = old_ppath

        old_home = new_env.pop('_OLD_VIRTUAL_PYTHONHOME', None)
        if old_home:
            new_env['PYTHON_HOME'] = old_home

        old_ps = new_env.pop('_OLD_VIRTUAL_PS1', None)
        if old_ps:
            new_env['PS1'] = old_ps

        new_env.pop('VIRTUAL_ENV', None)

        # Remove the configured set of environment variables from the
        # environment that we pass to Popen.
        for e in list(new_env.keys()):
            env_removables = list()
            if self._task_launcher:
                env_removables += self._task_launcher.env_removables
            for r in  env_removables:
                if e.startswith(r):
                    new_env.pop(e, None)

        return new_env


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        if not isinstance(tasks, list):
            tasks = [tasks]

        self.advance(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        for task in tasks:
            self._handle_task(task)


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        # prep stdout/err so that we can append w/o checking for None
        task['stdout'] = ''
        task['stderr'] = ''

        if not self._orte_initialized:
            self._log.debug("ORTE not yet initialized!")
            ret = self.init_orte(task)
            if ret != 0:
                self._log.debug("ORTE initialisation failed!")
            else:
                self._log.debug("ORTE initialisation succeeded!")

        try:
            launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher")

            self._log.debug("Launching task with %s (%s).",
                            launcher.name, launcher.launch_command)

            assert(task['slots']), 'task unscheduled'
            # Start a new subprocess to launch the task
            self.spawn(launcher=launcher, task=task)

        except Exception as e:
            # append the startup error to the tasks stderr.  This is
            # not completely correct (as this text is not produced
            # by the task), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running Task: %s", str(e))
            task['stderr'] += "\nPilot cannot start task:\n%s\n%s" \
                            % (str(e), traceback.format_exc())

            # Free the Slots, Flee the Flots, Ree the Frots!
            if task['slots']:
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

            self.advance(task, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def task_spawned_cb(self, task, status):

        with self.task_map_lock:
            task = self.task_map[task]
        uid = task['uid']

        if status:
            with self.task_map_lock:
                del self.task_map[task]

            # task launch failed
            self._prof.prof('exec_fail', uid=uid)
            self._log.error("task %s startup failed: %s", uid, status)
            self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

            task['target_state'] = rps.FAILED
            self.advance(task, rps.AGENT_STAGING_OUTPUT_PENDING,
                         publish=True, push=True)

        else:
            task['started'] = time.time()
            self._log.debug("task %s startup ok", uid)
            self.advance(task, rps.AGENT_EXECUTING, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def task_completed_cb(self, task, exit_code):

        timestamp = time.time()

        with self.task_map_lock:
            task = self.task_map[task]
            del self.task_map[task]

        self._prof.prof('exec_stop', uid=task['uid'])
        self._log.info("task %s finished [%s]", task['uid'], exit_code)

        task['exit_code'] = exit_code
        task['finished']  = timestamp

        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

        if exit_code != 0:
            # task failed - fail after staging output
            task['target_state'] = rps.FAILED

        else:
            # task finished cleanly.  We always move to stageout, even if there
            # are no staging directives -- at the very least, we'll upload
            # stdout/stderr
            task['target_state'] = rps.DONE

        self.advance(task, rps.AGENT_STAGING_OUTPUT_PENDING,
                     publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def init_orte(self, task):

        # FIXME: it feels as a hack to get the DVM URI from the Task

        slots = task['slots']

        if 'lm_info' not in slots:
            raise RuntimeError('No lm_info to init via %s: %s'
                               % (self.name, slots))

        if not slots['lm_info']:
            raise RuntimeError('lm_info missing for %s: %s'
                               % (self.name, slots))

        if 'dvm_uri' not in slots['lm_info']:
            raise RuntimeError('dvm_uri not in lm_info for %s: %s'
                               % (self.name, slots))

        dvm_uri = slots['lm_info']['dvm_uri']

        # Notify orte that we are using threads and that we require mutexes
        orte_lib.opal_set_using_threads(True)

        argv_keepalive = [
            ffi.new("char[]", "RADICAL-Pilot"),  # will be stripped off by lib
            ffi.new("char[]", "--hnp"), ffi.new("char[]", str(dvm_uri)),
            ffi.NULL,  # required
        ]
        argv = ffi.new("char *[]", argv_keepalive)
        ret = orte_lib.orte_submit_init(3, argv, ffi.NULL)

        self._myhandle = ffi.new_handle(self)
        self._orte_initialized = True

        return ret


    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, task):

        sandbox = task['task_sandbox_path']

        if False:
            task_tmpdir = '%s/%s' % (self.tmpdir, task['uid'])
        else:
            task_tmpdir = sandbox

        rec_makedir(task_tmpdir)

        # TODO: pre_exec
        # # Before the Big Bang there was nothing
        # if task['description']['pre_exec']:
        #     fail = ' (echo "pre_exec failed"; false) || exit'
        #     pre  = ''
        #     for elem in task['description']['pre_exec']:
        #         pre += "%s || %s\n" % (elem, fail)
        #     # Note: extra spaces below are for visual alignment
        #     launch_script.write("# Pre-exec commands\n")
        #     if self._prof.enabled:
        #         launch_script.write("echo task_pre_start `%s` >> %s/%s.prof\n"\
        #                           % (task['gtod'], sandbox, task['uid']))
        #     launch_script.write(pre)
        #     if self._prof.enabled:
        #         launch_script.write("echo task_pre_stop `%s` >> %s/%s.prof\n" \
        #                           % (task['gtod'], sandbox, task['uid']))

        # TODO: post_exec
        # # After the universe dies the infrared death, there will be nothing
        # if task['description']['post_exec']:
        #     fail = ' (echo "post_exec failed"; false) || exit'
        #     post = ''
        #     for elem in task['description']['post_exec']:
        #         post += "%s || %s\n" % (elem, fail)
        #     launch_script.write("# Post-exec commands\n")
        #     if self._prof.enabled:
        #         launch_script.write("echo task_post_start `%s` >> %s/%s.prof\n"
        #                           % (task['gtod'], sandbox, task['uid']))
        #     launch_script.write('%s\n' % post)
        #     if self._prof.enabled:
        #         launch_script.write("echo task_post_stop  `%s` >> %s/%s.prof\n"
        #                           % (task['gtod'], sandbox, task['uid']))


        # The actual command line, constructed per launch-method
        try:
            orte_command, task_command = launcher.construct_command(task, None)
        except Exception as e:
            msg = "Error in spawner (%s)" % e
            self._log.exception(msg)
            raise RuntimeError(msg)

        # Construct arguments to submit_job
        arg_list = []

        # Take the orte specific commands and split them
        for arg in orte_command.split():
            arg_list.append(ffi.new("char[]", str(arg)))

        # Set the working directory
        arg_list.append(ffi.new("char[]", "--wdir"))
        arg_list.append(ffi.new("char[]", str(sandbox)))

        # Set RP environment variables
        rp_envs = [
            "RP_SESSION_ID=%s"    % self._cfg['sid'],
            "RP_PILOT_ID=%s"      % self._cfg['pid'],
            "RP_AGENT_ID=%s"      % self._cfg['aid'],
            "RP_SPAWNER_ID=%s"    % self.uid,
            "RP_TASK_ID=%s"       % task['uid'],
            "RP_TASK_NAME=%s"     % task['description'].get('name'),
            "RP_PILOT_STAGING=%s" % self._pwd
        ]
        for env in rp_envs:
            arg_list.append(ffi.new("char[]", "-x"))
            arg_list.append(ffi.new("char[]", str(env)))

        # Set pre-populated environment variables
        if self._task_environment:
            for key,val in self._task_environment.items():
                arg_list.append(ffi.new("char[]", "-x"))
                arg_list.append(ffi.new("char[]", "%s=%s" % (key, val)))

        # Set environment variables specified for this Task
        if task['description']['environment']:
            for key,val in task['description']['environment'].items():
                arg_list.append(ffi.new("char[]", "-x"))
                arg_list.append(ffi.new("char[]", "%s=%s" % (key, val)))

        # Let the orted write stdout and stderr to rank-based output files
        arg_list.append(ffi.new("char[]", "--output-filename"))
        arg_list.append(ffi.new("char[]", "%s:nojobid,nocopy" % str(sandbox)))

        # Save retval of actual Task application (in case we have post-exec)
        task_command += "; RETVAL=$?"

        # Wrap in (sub)shell for output redirection
        arg_list.append(ffi.new("char[]", "sh"))
        arg_list.append(ffi.new("char[]", "-c"))
        if self._prof.enabled:
            task_command = "echo script task_start `%s` >> %s/%s.prof; " \
                         % (self.gtod, sandbox, task['uid']) \
                         + "echo script task_exec_start `%s` >> %s/%s.prof; " \
                         % (self.gtod, sandbox, task['uid']) \
                         + task_command \
                         + "; echo script task_exec_stop `%s` >> %s/%s.prof" \
                         % (self.gtod, sandbox, task['uid'])
        arg_list.append(ffi.new("char[]", str("%s; exit $RETVAL"
                                            % str(task_command))))

        self._log.debug("Launching task %s via %s %s", task['uid'],
                        orte_command, task_command)

        # NULL termination, required by ORTE
        arg_list.append(ffi.NULL)
        argv = ffi.new("char *[]", arg_list)

        # stdout/stderr filenames can'task be set with orte
        # TODO: assert here or earlier?
        # assert task['description'].get('stdout') == None
        # assert task['description'].get('stderr') == None

        # prepare stdout/stderr
        # TODO: when mpi==True && cores>1 there will be multiple files that need
        #       to be concatenated.
        task['stdout_file'] = os.path.join(sandbox, 'rank.0/stdout')
        task['stderr_file'] = os.path.join(sandbox, 'rank.0/stderr')

        # Submit to the DVM!
        index = ffi.new("int *")
        with self.task_map_lock:

            self._prof.prof('exec_start', uid=task['uid'])
            rc = orte_lib.orte_submit_job(argv, index, orte_lib.launch_cb,
                                          self._myhandle, orte_lib.finish_cb,
                                          self._myhandle)
            if rc:
                raise Exception("submit job failed with error: %d" % rc)

            self.task_map[index[0]] = task      # map ORTE index to Task
            self._prof.prof('exec_ok', uid=task['uid'])

        self._log.debug("Task %d submitted!", task['uid'])


# ------------------------------------------------------------------------------





import os
import time
import errno

import radical.utils as ru

# max number of t out/err chars to push to tail
MAX_IO_LOGLENGTH = 1024


# ------------------------------------------------------------------------------
#
def tail(txt, maxlen=MAX_IO_LOGLENGTH):

    # shorten the given string to the last <n> characters, and prepend
    # a notification.  This is used to keep logging information in mongodb
    # manageable(the size of mongodb documents is limited).

    if not txt:
        return txt

    if len(txt) > maxlen:
        return "[... CONTENT SHORTENED ...]\n%s" % txt[-maxlen:]
    else:
        return txt


# ------------------------------------------------------------------------------
#
def get_rusage():

    import resource

    self_usage  = resource.getrusage(resource.RUSAGE_SELF)
    child_usage = resource.getrusage(resource.RUSAGE_CHILDREN)

    rtime = time.time()
    utime = self_usage.ru_utime  + child_usage.ru_utime
    stime = self_usage.ru_stime  + child_usage.ru_stime
    rss   = self_usage.ru_maxrss + child_usage.ru_maxrss

    return "real %3f sec | user %.3f sec | system %.3f sec | mem %.2f kB" \
         % (rtime, utime, stime, rss)


# ------------------------------------------------------------------------------
#
def rec_makedir(target):

    # recursive makedir which ignores errors if dir already exists

    try:
        os.makedirs(target)

    except OSError as e:
        # ignore failure on existing directory
        if e.errno == errno.EEXIST and os.path.isdir(os.path.dirname(target)):
            pass
        else:
            raise


# ----------------------------------------------------------------------------------
#
def create_tar(tgt, dnames):
    '''
    Create a tarball on the file system which contains all given directories
    '''

    uid   = os.getuid()
    gid   = os.getgid()
    mode  = 16893
    mtime = int(time.time())

    fout  = open(tgt, 'wb')

    def rpad(s, size):
        return s + (size - len(s)) * '\0'

    def write_dir(path):
        data  = rpad(path, 100) \
              + rpad('%o' % mode,   8) \
              + rpad('%o' % uid,    8) \
              + rpad('%o' % gid,    8) \
              + rpad('%o' % 0,     12) \
              + rpad('%o' % mtime, 12) \
              + 8 * '\0' + '5'
        cksum = 256 + sum(ord(h) for h in data)
        data  = rpad(data  , 512)
        data  = data  [:-364] + '%06o\0' % cksum + data[-357:]
        fout.write(ru.as_bytes(data))

    for dname in dnames:
        write_dir(dname)

    fout.close()


# ----------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
# READ the RADICAL-Pilot documentation: https://radicalpilot.readthedocs.io/
#
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if   len(sys.argv)  > 2: report.exit('Usage:\t%s [resource]\n\n' % sys.argv[0])
    elif len(sys.argv) == 2: resource = sys.argv[1]
    else                   : resource = 'local.localhost'

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/config.json' % os.path.dirname(os.path.abspath(__file__)))
        report.ok('>>ok\n')

        report.header('submit pilots')

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager(session=session)

        # Define an [n]-core local pilot that runs for [x] minutes
        # Here we use a dict to initialize the description object
        pd_init = {
                   'resource'      : resource,
                   'runtime'       : 15,  # pilot runtime (min)
                   'exit_on_error' : True,
                   'project'       : config[resource].get('project', None),
                   'queue'         : config[resource].get('queue', None),
                   'access_schema' : config[resource].get('schema', None),
                   'cores'         : config[resource].get('cores', 1),
                   'gpus'          : config[resource].get('gpus', 0),
                  }
        pdesc = rp.PilotDescription(pd_init)

        # Launch the pilot.
        pilot = pmgr.submit_pilots(pdesc)


        report.header('submit tasks')

        # Register the Pilot in a TaskManager object.
        tmgr = rp.TaskManager(session=session)
        tmgr.add_pilots(pilot)

        # Create a workload of char-counting a simple file.  We first create the
        # file right here, and then use it as task input data for each task.
        os.system('hostname >  input.dat')
        os.system('date     >> input.dat')
        os.system('hostname >  input2.dat')
        os.system('date     >> input2.dat')

        n = 16   # number of tasks to run
        report.info('create %d task description(s)\n\t' % n)

        tds = list()
        for i in range(0, n):

            # create a new Task description, and fill it.
            # Here we don't use dict initialization.
            td = rp.TaskDescription()
            td.executable     = 'ls'
            td.arguments      = ['-la']

            # Move multiple files to a Task by using a tarball
            td.input_staging  = [{'source': 'client:///input.dat',
                                   'target': 'task:///input.dat',
                                   'action': rp.TARBALL},
                                  {'source': 'client:///input2.dat',
                                   'target': 'task:///input2.dat',
                                   'action': rp.TARBALL}]

            tds.append(td)
            report.progress()
        report.ok('>>ok\n')

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        tasks = tmgr.submit_tasks(tds)

        # Wait for all tasks to reach a final state (DONE, CANCELED or FAILED).
        report.header('gather results')
        tmgr.wait_tasks()

        report.info('\n')
        for task in tasks:
            report.plain('  * %s: %s, exit: %3s, out: %s\n'
                        % (task.uid, task.state[:4],
                           task.exit_code, task.stdout.strip()[:35]))

        # delete the sample input files
        os.system('rm input.dat')


    except Exception as e:
        # Something unexpected happened in the pilot code above
        report.error('caught Exception: %s\n' % e)
        raise

    except (KeyboardInterrupt, SystemExit):
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        report.warn('exit requested\n')

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        session.close()

    report.header()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import copy
import stat
import time
import queue
import tempfile
import threading
import traceback
import subprocess

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc

from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
#
class ABDS(AgentExecutingComponent):

    # The name is rong based on the abstraction, but for the moment I do not
    # have any other ideas

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__ (self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        from .... import pilot as rp

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)

        self._cancel_lock     = ru.RLock()
        self._tasks_to_cancel = list()
        self._tasks_to_watch  = list()
        self._watch_queue     = queue.Queue()

        self._pid = self._cfg['pid']

        # run watcher thread
        self._terminate = threading.Event()
        self._watcher   = threading.Thread(target=self._watch, name="Watcher")
        self._watcher.daemon = True
        self._watcher.start ()

        # The AgentExecutingComponent needs the LaunchMethod to construct
        # commands.
        self._task_launcher = rp.agent.LaunchMethod.create(
                name    = self._cfg['task_launch_method'],
                cfg     = self._cfg,
                session = self._session)

        self._mpi_launcher = rp.agent.LaunchMethod.create(
                name    = self._cfg['mpi_launch_method'],
                cfg     = self._cfg,
                session = self._session)

        self._task_environment = self._populate_task_environment()

        self.gtod   = "%s/gtod" % self._pwd
        self.tmpdir = tempfile.gettempdir()

        # if we need to transplant any original env into the Task, we dig the
        # respective keys from the dump made by bootstrap_0.sh
        self._env_task_export = dict()
        if self._cfg.get('export_to_task'):
            with open('env.orig', 'r') as f:
                for line in f.readlines():
                    if '=' in line:
                        k,v = line.split('=', 1)
                        key = k.strip()
                        val = v.strip()
                        if key in self._cfg['export_to_task']:
                            self._env_task_export[key] = val


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        # terminate watcher thread
        self._terminate.set()
        if self._watcher:
            self._watcher.join()


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_tasks':

            self._log.info("cancel_tasks command (%s)" % arg)
            with self._cancel_lock:
                self._tasks_to_cancel.extend(arg['uids'])

        return True


    # --------------------------------------------------------------------------
    #
    def _populate_task_environment(self):
        """Derive the environment for the task's from our own environment."""

        # Get the environment of the agent
        new_env = copy.deepcopy(os.environ)

        #
        # Mimic what virtualenv's "deactivate" would do
        #
        old_path = new_env.pop('_OLD_VIRTUAL_PATH', None)
        if old_path:
            new_env['PATH'] = old_path

        old_ppath = new_env.pop('_OLD_VIRTUAL_PYTHONPATH', None)
        if old_ppath:
            new_env['PYTHONPATH'] = old_ppath

        old_home = new_env.pop('_OLD_VIRTUAL_PYTHONHOME', None)
        if old_home:
            new_env['PYTHON_HOME'] = old_home

        old_ps = new_env.pop('_OLD_VIRTUAL_PS1', None)
        if old_ps:
            new_env['PS1'] = old_ps

        new_env.pop('VIRTUAL_ENV', None)

        # Remove the configured set of environment variables from the
        # environment that we pass to Popen.
        for e in list(new_env.keys()):
            env_removables = list()
            if self._mpi_launcher : env_removables += self._mpi_launcher.env_removables
            if self._task_launcher: env_removables += self._task_launcher.env_removables
            for r in  env_removables:
                if e.startswith(r):
                    new_env.pop(e, None)

        return new_env


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        if not isinstance(tasks, list):
            tasks = [tasks]

        self.advance(tasks, rps.AGENT_SCHEDULING, publish=True, push=False)

        for task in tasks:
            self._handle_task(task)

        self.advance(tasks, rps.AGENT_EXECUTING_PENDING, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        try:
            if task['description']['mpi']:
                launcher = self._mpi_launcher
            else :
                launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher (mpi=%s)" % task['description']['mpi'])

            self._log.debug("Launching task with %s (%s).", launcher.name, launcher.launch_command)

            assert(task['slots']) # FIXME: no assert, but check
            self._prof.prof('exec', msg='task launch', uid=task['uid'])

            # Start a new subprocess to launch the task
            self.spawn(launcher=launcher, task=task)

        except Exception as e:
            # append the startup error to the tasks stderr.  This is
            # not completely correct (as this text is not produced
            # by the task), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running Task")
            task['stderr'] += "\nPilot cannot start task:\n%s\n%s" \
                            % (str(e), traceback.format_exc())

            # Free the Slots, Flee the Flots, Ree the Frots!
            if task['slots']:
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

            self.advance(task, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, task):

        self._prof.prof('spawn', msg='task spawn', uid=task['uid'])

        sandbox = task['task_sandbox_path']

        # make sure the sandbox exists
        rpu.rec_makedir(sandbox)

        # prep stdout/err so that we can append w/o checking for None
        task['stdout']  = ''
        task['stderr']  = ''

        launch_script_name = '%s/%s.sh' % (sandbox, task['uid'])

        with open(launch_script_name, "w") as launch_script:
            launch_script.write('#!/bin/sh\n\n')

            # Create string for environment variable setting
            env_string  = ''
            if task['description']['environment']:
                for key,val in task['description']['environment'].items():
                    env_string += 'export %s="%s"\n' % (key, val)
            env_string += 'export RP_SESSION_ID="%s"\n'    % self._cfg['sid']
            env_string += 'export RP_PILOT_ID="%s"\n'      % self._cfg['pid']
            env_string += 'export RP_AGENT_ID="%s"\n'      % self._cfg['aid']
            env_string += 'export RP_SPAWNER_ID="%s"\n'    % self.uid
            env_string += 'export RP_TASK_ID="%s"\n'       % task['uid']
            env_string += 'export RP_TASK_NAME="%s"\n'     % task['description'].get('name')
            env_string += 'export RP_GTOD="%s"\n'          % self.gtod
            env_string += 'export RP_PILOT_STAGING="%s"\n' % self._pwd
            if self._prof.enabled:
                env_string += 'export RP_PROF="%s/%s.prof"\n' % (sandbox, task['uid'])

            # also add any env vars requested for export by the resource config
            for k,v in self._env_task_export.items():
                env_string += "export %s=%s\n" % (k,v)

            env_string += '''
prof(){
    if test -z "$RP_PROF"
    then
        return
    fi
    event=$1
    now=$($RP_GTOD)
    echo "$now,$event,task_script,MainThread,$RP_TASK_ID,AGENT_EXECUTING," >> $RP_PROF
}
'''

            # also add any env vars requested in the task description
            if task['description']['environment']:
                for key,val in task['description']['environment'].items():
                    env_string += 'export %s=%s\n' % (key, val)

            launch_script.write('\n# Environment variables\n%s\n' % env_string)

            launch_script.write('prof task_start\n')
            launch_script.write('\n# Change to task sandbox\ncd %s\n' % sandbox)

            # Before the Big Bang there was nothing
            if task['description']['pre_exec']:
                pre_exec_string = ''
                if isinstance(task['description']['pre_exec'], list):
                    for elem in task['description']['pre_exec']:
                        pre_exec_string += "%s\n" % elem
                else:
                    pre_exec_string += "%s\n" % task['description']['pre_exec']
                # Note: extra spaces below are for visual alignment
                launch_script.write("\n# Pre-exec commands\n")
                launch_script.write('prof task_pre_start\n')
                launch_script.write(pre_exec_string)
                launch_script.write('prof task_pre_stop\n')

            # YARN pre execution folder permission change
            launch_script.write('\n## Changing Working Directory permissions for YARN\n')
            launch_script.write('old_perm="`stat -c %a .`"\n')
            launch_script.write('chmod -R 777 .\n')

            # The actual command line, constructed per launch-method
            try:
                self._log.debug("Launch Script Name %s",launch_script_name)
                launch_command, hop_cmd = launcher.construct_command(task, launch_script_name)
                self._log.debug("Launch Command %s from %s", launch_command, launcher.name)

                if hop_cmd : cmdline = hop_cmd
                else       : cmdline = launch_script_name

            except Exception as e:
                msg = "Error in spawner (%s)" % e
                self._log.exception(msg)
                raise RuntimeError(msg)

            launch_script.write("\n# The command to run\n")
            launch_script.write('prof task_exec_start\n')
            launch_script.write("%s\n" % launch_command)
            launch_script.write("RETVAL=$?\n")
            launch_script.write("\ncat Ystdout\n")
            launch_script.write('prof task_exec_stop\n')

            # After the universe dies the infrared death, there will be nothing
            if task['description']['post_exec']:
                post_exec_string = ''
                if isinstance(task['description']['post_exec'], list):
                    for elem in task['description']['post_exec']:
                        post_exec_string += "%s\n" % elem
                else:
                    post_exec_string += "%s\n" % task['description']['post_exec']
                launch_script.write("\n# Post-exec commands\n")
                launch_script.write('prof task_post_start\n')
                launch_script.write('%s\n' % post_exec_string)
                launch_script.write('prof task_post_stop\n')

            # YARN pre execution folder permission change
            launch_script.write('\n## Changing Working Directory permissions for YARN\n')
            launch_script.write('chmod $old_perm .\n')

            launch_script.write("\n# Exit the script with the return code from the command\n")
            launch_script.write("exit $RETVAL\n")

        # done writing to launch script, get it ready for execution.
        st = os.stat(launch_script_name)
        os.chmod(launch_script_name, st.st_mode | stat.S_IEXEC)
        self._prof.prof('control', msg='launch script constructed', uid=task['uid'])

        # prepare stdout/stderr
        stdout_file = task['description'].get('stdout') or '%s.out' % task['uid']
        stderr_file = task['description'].get('stderr') or '%s.err' % task['uid']

        task['stdout_file'] = os.path.join(sandbox, stdout_file)
        task['stderr_file'] = os.path.join(sandbox, stderr_file)

        _stdout_file_h = open(task['stdout_file'], "w+")
        _stderr_file_h = open(task['stderr_file'], "w+")
        self._prof.prof('control', msg='stdout and stderr files created', uid=task['uid'])

        self._log.info("Launching task %s via %s in %s", task['uid'], cmdline, sandbox)

        self._prof.prof('spawn', msg='spawning passed to popen', uid=task['uid'])

        task['proc'] = subprocess.Popen(args               = cmdline,
                                        bufsize            = 0,
                                        executable         = None,
                                        stdin              = None,
                                        stdout             = _stdout_file_h,
                                        stderr             = _stderr_file_h,
                                        preexec_fn         = None,
                                        close_fds          = True,
                                        shell              = True,
                                        cwd                = sandbox,
                                        env                = self._task_environment,
                                        universal_newlines = False,
                                        startupinfo        = None,
                                        creationflags      = 0)

        self._prof.prof('spawn', msg='spawning passed to popen', uid=task['uid'])
        self._watch_queue.put(task)


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        try:
            self._prof.prof('run', uid=self._pid)

            while not self._terminate.is_set():

                tasks = list()

                try:

                    # we don't want to only wait for one Task -- then we would
                    # pull Task state too frequently.  OTOH, we also don't want to
                    # learn about tasks until all slots are filled, because then
                    # we may not be able to catch finishing tasks in time -- so
                    # there is a fine balance here.  Balance means 100 (FIXME).
                  # self._prof.prof('pull')
                    MAX_QUEUE_BULKSIZE = 100
                    while len(tasks) < MAX_QUEUE_BULKSIZE :
                        tasks.append (self._watch_queue.get_nowait())

                except queue.Empty:

                    # nothing found -- no problem, see if any tasks finished
                    pass

                # add all tasks we found to the watchlist
                for task in tasks :

                    self._prof.prof('passed', msg="ExecWatcher picked up task",
                                              uid=task['uid'])
                    self._tasks_to_watch.append (task)

                # check on the known tasks.
                action = self._check_running()

                if not action and not tasks :
                    # nothing happened at all!  Zzz for a bit.
                    time.sleep(self._cfg['db_poll_sleeptime'])

        except Exception as e:
            self._log.exception("Error in ExecWorker watch loop (%s)" % e)
            # FIXME: this should signal the ExecWorker for shutdown...

        self._prof.prof('stop', uid=self._pid)
        self._prof.flush()


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self):

        action = 0

        for task in self._tasks_to_watch:

            sandbox = '%s/%s' % (self._pwd, task['uid'])

            # ------------------------------------------------------------------
            # This code snippet reads the YARN application report file and if
            # the application is RUNNING it update the state of the Task with
            # the right time stamp. In any other case it works as it was.
            logfile = '%s/%s' % (sandbox, '/YarnApplicationReport.log')
            if task['state'] == rps.AGENT_EXECUTING_PENDING \
                    and os.path.isfile(logfile):

                yarnreport = open(logfile,'r')
                report_contents = yarnreport.readlines()
                yarnreport.close()

                for report_line in report_contents:
                    if report_line.find('RUNNING') != -1:
                        self._log.debug(report_contents)
                        line = report_line.split(',')
                        ts   = (int(line[3].split('=')[1])/1000)
                        action += 1
                        proc = task['proc']
                        self._log.debug('Proc Print {0}'.format(proc))
                        del(task['proc'])  # proc is not json serializable
                        self.advance(task, rps.AGENT_EXECUTING, publish=True,
                                     push=False,ts=ts)
                        task['proc'] = proc

                        # FIXME: Ioannis, what is this supposed to do?  I wanted
                        # to update the state of the task but keep it in the
                        # watching queue. I am not sure it is needed anymore.
                        index = self._tasks_to_watch.index(task)
                        self._tasks_to_watch[index] = task

            else :
                # poll subprocess object
                exit_code = task['proc'].poll()

                if exit_code is None:
                    # Process is still running

                    if task['uid'] in self._tasks_to_cancel:

                        # FIXME: there is a race condition between the state poll
                        # above and the kill command below.  We probably should pull
                        # state after kill again?

                        # We got a request to cancel this task
                        action += 1
                        task['proc'].kill()
                        task['proc'].wait() # make sure proc is collected

                        with self._cancel_lock:
                            self._tasks_to_cancel.remove(task['uid'])

                        self._prof.prof('final', msg="execution canceled",
                                uid=task['uid'])

                        self._tasks_to_watch.remove(task)

                        del(task['proc'])  # proc is not json serializable
                        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)
                        self.advance(task, rps.CANCELED, publish=True, push=False)

                else:
                    self._prof.prof('exec', msg='execution complete', uid=task['uid'])


                    # make sure proc is collected
                    task['proc'].wait()

                    # we have a valid return code -- task is final
                    action += 1
                    self._log.info("Task %s has return code %s.", task['uid'], exit_code)

                    task['exit_code'] = exit_code

                    # Free the Slots, Flee the Flots, Ree the Frots!
                    self._tasks_to_watch.remove(task)
                    del(task['proc'])  # proc is not json serializable
                    self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                    if exit_code != 0:
                        # The task failed - fail after staging output
                        self._prof.prof('final', msg="execution failed",
                                uid=task['uid'])
                        task['target_state'] = rps.FAILED

                    else:
                        # The task finished cleanly, see if we need to deal with
                        # output data.  We always move to stageout, even if there are no
                        # directives -- at the very least, we'll upload stdout/stderr
                        self._prof.prof('final', msg="execution succeeded",
                                uid=task['uid'])
                        task['target_state'] = rps.DONE

                    self.advance(task, rps.AGENT_STAGING_OUTPUT_PENDING, publish=True, push=True)

        return action






__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class RSH(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        # Instruct the ExecWorkers to unset this environment variable.
        # Otherwise this will break nested RSH with SHELL spawner, i.e. when
        # both the sub-agent and CUs are started using RSH.
        self.env_removables.extend(["RP_SPAWNER_HOP"])


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # Find rsh command
        self.launch_command = ru.which('rsh')

        if not self.launch_command:
            raise RuntimeError("rsh not found!")


    # --------------------------------------------------------------------------
    #
    def construct_command(self, t, launch_script_hop):

        slots        = t['slots']
        td          = t['description']
        task_exec    = td['executable']
        task_env     = td.get('environment') or dict()
        task_args    = td.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        if not launch_script_hop:
            raise ValueError("RSH launch method needs launch_script_hop!")

        if 'nodes' not in slots:
            raise RuntimeError('insufficient information to launch via %s: %s'
                              % (self.name, slots))

        if len(slots['nodes']) > 1:
            raise RuntimeError('rsh cannot run multinode tasks')

        host = slots['nodes'][0]['name']

        # Pass configured and available environment variables to the remote shell
        export_vars  = ' '.join(['%s=%s' % (var, os.environ[var])
                                 for var in self.EXPORT_ENV_VARIABLES
                                  if var in os.environ])
        export_vars += ' '.join(['%s=%s' % (var, task_env[var])
                                 for var in task_env])

        # Command line to execute launch script via rsh on host
        rsh_hop_cmd = "%s %s %s %s" % (self.launch_command, host,
                                       export_vars, launch_script_hop)

        return task_command, rsh_hop_cmd


# ------------------------------------------------------------------------------





import time

import radical.utils as ru


# ------------------------------------------------------------------------------
#
class Request(object):

    # poor man's future
    # TODO : use proper future implementation
    # FIXME: use ru.Description base class


    # --------------------------------------------------------------------------
    #
    def __init__(self, req):

        self._req     = req
        self._state   = 'NEW'
        self._result  = None

        self._uid = self._req['uid']
        if not self._uid:
            self._uid = ru.generate_id('request')


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    @property
    def state(self):
        return self._state


    @property
    def work(self):
        return self._req


    @property
    def result(self):
        return self._result


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        '''
        produce the request message to be sent over the wire to the workers
        '''

        # FIXME: we should not need to reconstruict the dict
        return {'uid'      : self._uid,
                'state'    : self._state,
                'cpus'     : self._req.get('cores', 1),
                'gpus'     : self._req.get('gpus',  0),
                'timeout'  : self._req.get('timeout'),
                'mode'     : self._req['mode'],
                'data'     : self._req['data'],
                'result'   : self._result}


    # --------------------------------------------------------------------------
    #
    def set_result(self, out, err, ret):
        '''
        This is called by the master to fulfill the future
        '''

        self._result = {'out': out,
                        'err': err,
                        'ret': ret}

        if ret: self._state = 'FAILED'
        else  : self._state = 'DONE'


    # --------------------------------------------------------------------------
    #
    def wait(self):

        while self.state not in ['DONE', 'FAILED']:
            time.sleep(0.1)

        return self._result


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2020-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import math
import os
import signal
import time

import subprocess    as mp
import threading     as mt

import radical.utils as ru

from .base import LaunchMethod

PTL_MAX_MSG_SIZE = 1024 * 1024 * 1024 * 1

DVM_URI_FILE_TPL   = '%(base_path)s/prrte.%(dvm_id)03d.uri'
DVM_HOSTS_FILE_TPL = '%(base_path)s/prrte.%(dvm_id)03d.hosts'


# ------------------------------------------------------------------------------
#
class PRTE2(LaunchMethod):

    """
    PMIx Reference Run Time Environment v2
    (https://github.com/openpmix/prrte)
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        # We remove all PRUN related environment variables from the launcher
        # environment, so that we can use PRUN for both launch of the
        # (sub-)agent and Task execution.
        self.env_removables.extend(['OMPI_', 'OPAL_', 'PMIX_'])

        self._verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))

    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        return "echo $PMIX_RANK"

    # --------------------------------------------------------------------------
    #
    def _configure(self):

        if not ru.which('prun'):
            raise Exception('prun command not found')

        self.launch_command = 'prun'

    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler):

        prte_cmd = ru.which('prte')
        if not prte_cmd:
            raise Exception('prte command not found')

        prte_prefix = os.environ.get('UMS_OMPIX_PRRTE_DIR') or \
                      os.environ.get('PRRTE_DIR', '')

        prte_info = {}
        # get OpenRTE/PRTE version
        out, _, _ = ru.sh_callout('prte_info | grep "RTE"', shell=True)
        for line in out.split('\n'):

            line = line.strip()

            if not prte_info.get('name'):
                if 'Open RTE' in line:
                    prte_info['name'] = 'Open RTE'
                elif 'PRTE' in line:
                    prte_info['name'] = 'PRTE'

            if 'RTE:' in line:
                prte_info['version'] = line.split(':')[1].strip()
            elif 'RTE repo revision:' in line:
                prte_info['version_detail'] = line.split(':')[1].strip()

        if prte_info.get('name'):
            log.info('version of %s: %s [%s]',
                     prte_info['name'],
                     prte_info.get('version'),
                     prte_info.get('version_detail'))
        else:
            log.info('version of OpenRTE/PRTE not determined')

        # get `dvm_count` from resource config
        # FIXME: another option is to introduce `nodes_per_dvm` in config
        #        dvm_count = math.ceil(len(rm.node_list) / nodes_per_dvm)
        dvm_count      = cfg['resource_cfg'].get('dvm_count') or 1
        nodes_per_dvm  = math.ceil(len(rm.node_list) / dvm_count)

        # additional info to form prrte related files (uri-/hosts-file)
        dvm_file_info  = {'base_path': os.getcwd()}
        # list of threading.Event that checks DVM's readiness
        dvm_ready_list = []

        # ----------------------------------------------------------------------
        def _watch_dvm(dvm_process, dvm_id, dvm_ready_flag):

            log.info('prte-%s watcher started', dvm_id)

            retval = dvm_process.poll()
            while retval is None:
                line = dvm_process.stdout.readline().strip()
                if line:
                    log.debug('prte-%s output: %s', dvm_id, line)

                    # final confirmation that DVM has started successfully
                    if 'DVM ready' in str(line):
                        log.info('prte-%s is ready', dvm_id)
                        dvm_ready_flag.set()

                else:
                    time.sleep(1.)

            if retval != 0:
                # send a kill signal to the main thread.
                # We know that Python and threading are likely not to play well
                # with signals - but this is an exceptional case, and not part
                # of the standard termination sequence. If the signal is
                # swallowed, the next prun-call will trigger termination anyway.
                os.kill(os.getpid(), signal.SIGKILL)
                raise RuntimeError('PRTE DVM died (dvm_id=%s)' % dvm_id)

            log.info('prte-%s watcher stopped', dvm_id)

        # ----------------------------------------------------------------------
        def _start_dvm(dvm_id, dvm_size, dvm_ready_flag):

            file_info = dict(dvm_file_info)
            file_info['dvm_id'] = dvm_id

            prte  = '%s'               % prte_cmd
            prte += ' --prefix %s'     % prte_prefix
            prte += ' --report-uri %s' % DVM_URI_FILE_TPL   % file_info
            prte += ' --hostfile %s'   % DVM_HOSTS_FILE_TPL % file_info

            # large tasks imply large message sizes, we need to account for that
            # FIXME: need to derive the message size from DVM size - smaller
            #        DVMs will never need large messages, as they can't run
            #        large tasks
            prte += ' --pmixmca ptl_base_max_msg_size %d'  % PTL_MAX_MSG_SIZE
            prte += ' --prtemca routed_radix %d' % dvm_size
            # 2 tweaks on Summit, which should not be needed in the long run:
            # - ensure 1 ssh per dvm
            prte += ' --prtemca plm_rsh_num_concurrent %d' % dvm_size
            # - avoid 64 node limit (ssh connection limit)
            prte += ' --prtemca plm_rsh_no_tree_spawn 1'

            # to select a set of hardware threads that should be used for
            # processing the following option should set hwthread IDs:
            #   --prtemca hwloc_default_cpu_list "0-83,88-171"

            # Use (g)stdbuf to disable buffering.  We need this to get the
            # "DVM ready" message to ensure DVM startup completion
            #
            # The command seems to be generally available on Cray's,
            # if not, we can code some home-cooked pty stuff (TODO)
            stdbuf_cmd = ru.which(['stdbuf', 'gstdbuf'])
            if not stdbuf_cmd:
                raise Exception('(g)stdbuf command not found')
            stdbuf_arg = '-oL'

            # base command = (g)stdbuf <args> + prte <args>
            cmd = '%s %s %s ' % (stdbuf_cmd, stdbuf_arg, prte)

            # additional (debug) arguments to prte
            verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))
            if verbose:
                cmd += ' '.join(['--prtemca plm_base_verbose 5'])

            cmd = cmd.strip()

            log.info('start prte-%s on %d nodes [%s]', dvm_id, dvm_size, cmd)
            profiler.prof(event='dvm_start',
                          uid=cfg['pid'],
                          msg='dvm_id=%s' % dvm_id)

            dvm_process = mp.Popen(cmd.split(),
                                   stdout=mp.PIPE,
                                   stderr=mp.STDOUT)
            dvm_watcher = mt.Thread(target=_watch_dvm,
                                    args=(dvm_process, dvm_id, dvm_ready_flag),
                                    daemon=True)
            dvm_watcher.start()

            dvm_uri = None
            for _ in range(100):
                time.sleep(.5)

                try:
                    with open(DVM_URI_FILE_TPL % file_info, 'r') as fin:
                        for line in fin.readlines():
                            if '://' in line:
                                dvm_uri = line.strip()
                                break
                except Exception as e:
                    log.debug('DVM URI file missing: %s...', str(e)[:24])

                if dvm_uri:
                    break

            if not dvm_uri:
                raise Exception('DVM URI not found')

            log.info('prte-%s startup successful: [%s]', dvm_id, dvm_uri)
            profiler.prof(event='dvm_ok',
                          uid=cfg['pid'],
                          msg='dvm_id=%s' % dvm_id)

            return dvm_uri

        # ----------------------------------------------------------------------

        # format: {<id>: {'nodes': [...], 'details': {'dvm_uri': <uri>}}}
        # LM specific parameters will be kept under `'details'`
        partitions = {}

        # go through every dvm instance
        for _dvm_id in range(dvm_count):

            node_list = rm.node_list[_dvm_id       * nodes_per_dvm:
                                     (_dvm_id + 1) * nodes_per_dvm]
            dvm_file_info.update({'dvm_id': _dvm_id})
            # write hosts file
            with open(DVM_HOSTS_FILE_TPL % dvm_file_info, 'w') as fout:
                num_slots = rm.cores_per_node * rm.smt
                for node in node_list:
                    fout.write('%s slots=%d\n' % (node[0], num_slots))
            vm_size = len(node_list)

            dvm_ready = mt.Event()
            dvm_ready_list.append(dvm_ready)

            partitions[str(_dvm_id)] = {
                'nodes'  : [node[1] for node in node_list],
                'details': {'dvm_uri': _start_dvm(_dvm_id, vm_size, dvm_ready)}}

        # check that all DVMs are ready
        for _dvm_id, _dvm_ready in enumerate(dvm_ready_list):

            if not _dvm_ready.is_set():

                partition_id = str(_dvm_id)

                # extra time to confirm that "DVM ready" was just delayed
                if _dvm_ready.wait(timeout=10.):
                    continue

                log.info('prte-%s to be re-started', _dvm_id)

                # terminate DVM (if it is still running)
                _dvm_uri = partitions[partition_id]['details']['dvm_uri']
                try   : ru.sh_callout('pterm --dvm-uri "%s"' % _dvm_uri)
                except: pass

                # re-start DVM
                vm_size = len(partitions[partition_id]['nodes'])
                partitions[partition_id]['details']['dvm_uri'] = \
                    _start_dvm(_dvm_id, vm_size, _dvm_ready)

                # FIXME: with the current approach there is only one attempt to
                #        restart DVM(s). If a failure during the start process
                #        will keep appears then need to consider re-assignment
                #        of nodes from failed DVM(s) + add time for re-started
                #        DVM to stabilize: `time.sleep(10.)`

        lm_info = {'partitions'  : partitions,
                   'version_info': prte_info,
                   'cvd_id_mode' : 'physical'}

        # we need to inform the actual LaunchMethod instance about partitions,
        # which includes list of nodes per partition/dvm and dvm_uri (which is
        # in section 'details'). So we pass it back to the ResourceManager,
        # which will keep it in an 'lm_info', which will then be passed as part
        # of the slots via the scheduler, as well partition 'nodes' will be kept
        # in the ResourceManager separately.
        return lm_info

    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_shutdown_hook(cls, name, cfg, rm, lm_info, log, profiler):
        """
        This hook is symmetric to the hook `rm_config_hook`, it is called
        during shutdown sequence, for the sake of freeing allocated resources.
        """
        cmd = ru.which('pterm')
        if not cmd:
            raise Exception('termination command not found')

        for p_id, p_data in lm_info.get('partitions', {}).items():
            dvm_uri = p_data['details']['dvm_uri']
            try:
                log.info('terminating prte-%s (%s)', p_id, dvm_uri)
                _, err, _ = ru.sh_callout('%s --dvm-uri "%s"' % (cmd, dvm_uri))
                log.debug('termination status: %s', err.strip('\n') or '-')
                profiler.prof(event='dvm_stop',
                              uid=cfg['pid'],
                              msg='dvm_id=%s' % p_id)

            except Exception as e:
                # use the same event name as for runtime failures - those are
                # not distinguishable at the moment from termination failures
                log.debug('prte-%s termination failed (%s)', p_id, dvm_uri)
                profiler.prof(event='dvm_fail',
                              uid=cfg['pid'],
                              msg='dvm_id=%s | error: %s' % (p_id, e))

    # --------------------------------------------------------------------------
    #
    def construct_command(self, t, launch_script_hop):

        # latest implementation of PRRTE is able to handle bulk submissions
        # of prun-commands, but in case errors will come out then the delay
        # should be set: `time.sleep(.1)`

        slots     = t['slots']
        td        = t['description']

        task_exec = td['executable']
        task_args = td.get('arguments') or []

        n_procs   = td.get('cpu_processes') or 1
        n_threads = td.get('cpu_threads')   or 1

        if not slots.get('lm_info'):
            raise RuntimeError('lm_info not set (%s): %s' % (self.name, slots))

        if not slots['lm_info'].get('partitions'):
            raise RuntimeError('no partitions (%s): %s' % (self.name, slots))

        flags  = ' --np %d'                               % n_procs
        flags += ' --map-by hwthread:PE=%d:OVERSUBSCRIBE' % n_threads
        flags += ' --bind-to hwthread:overload-allowed'
        if self._verbose:
            flags += ':REPORT'

        dvm_uri = ''
        if 'nodes' not in slots:
            # this task is unscheduled - we leave it to PRRTE/PMI-X
            # to correctly place the task
            pass
        else:
            # FIXME: ensure correct binding for procs and threads via slotfile

            hosts = [node['name'] for node in slots['nodes']]
            flags += ' --host %s:%s' % (hosts[0], len(hosts))

            # scheduler makes sure that all nodes are from the same dvm
            _host_uid = slots['nodes'][0]['uid']
            for partition in slots['lm_info']['partitions'].values():
                if _host_uid in partition['nodes']:
                    dvm_uri = '--dvm-uri "%s"' % partition['details']['dvm_uri']
                    break

        flags += ' --pmixmca ptl_base_max_msg_size %d' % PTL_MAX_MSG_SIZE
        flags += ' --verbose'  # needed to get prte profile events

        task_args_str = self._create_arg_string(task_args)
        if task_args_str:
            task_exec += ' %s' % task_args_str

        cmd = '%s %s %s %s' % (self.launch_command, dvm_uri, flags, task_exec)

        return cmd, None

# ------------------------------------------------------------------------------




__copyright__ = 'Copyright 2016-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import logging
import math
import os
import signal
import time

import threading     as mt
import subprocess    as mp

import radical.utils as ru

from .base import LaunchMethod

PTL_BASE_MAX_MSG_SIZE = 1024 * 1024 * 1024 * 1


# ------------------------------------------------------------------------------
#
class PRTE(LaunchMethod):


    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        # We remove all PRUN related environment variables from the launcher
        # environment, so that we can use PRUN for both launch of the
        # (sub-)agent and Task execution.
        self.env_removables.extend(["OMPI_", "OPAL_", "PMIX_"])

        self._verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        return "echo $PMIX_RANK"


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _watch_dvm(dvm_process, dvm_id, log):

        log.info('start prte-%s watcher', dvm_id)

        retval = dvm_process.poll()
        while retval is None:
            line = dvm_process.stdout.readline().strip()
            if line:
                log.debug('prte-%s output: %s', dvm_id, line)
            else:
                time.sleep(1.)

        if retval != 0:
            # send a kill signal to the main thread.
            # We know that Python and threading are likely not to play well
            # with signals - but this is an exceptional case, and not part
            # of the standard termination sequence.  If the signal is
            # swallowed, the next `prun` call will trigger
            # termination anyway.
            os.kill(os.getpid(), signal.SIGKILL)
            raise RuntimeError('PRTE DVM died')

        log.info('prte-%s watcher stopped (%d)', dvm_id, dvm_process.returncode)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler):

        prte_cmd = ru.which('prte')
        if not prte_cmd:
            raise Exception('prte command not found')

        # get OpenRTE/PRTE version
        out, _, _ = ru.sh_callout('prte_info | grep "RTE"', shell=True)
        prte_info = {}
        for line in out.split('\n'):

            line = line.strip()

            if not prte_info.get('name'):
                if 'Open RTE' in line:
                    prte_info['name'] = 'Open RTE'
                elif 'PRTE' in line:
                    prte_info['name'] = 'PRTE'

            if 'RTE:' in line:
                prte_info['version'] = line.split(':')[1].strip()
            elif 'RTE repo revision:' in line:
                prte_info['version_detail'] = line.split(':')[1].strip()

        if prte_info.get('name'):
            log.info('version of %s: %s [%s]',
                     prte_info['name'],
                     prte_info.get('version'),
                     prte_info.get('version_detail'))
        else:
            log.info('version of OpenRTE/PRTE not determined')

        # get `dvm_count` from resource config
        # FIXME: another option is to introduce `nodes_per_dvm` in config
        #        dvm_count = math.ceil(num_nodes / nodes_per_dvm)
        num_nodes      = len(rm.node_list)
        dvm_count      = cfg['resource_cfg'].get('dvm_count') or 1
        nodes_per_dvm  = math.ceil(num_nodes / dvm_count)

        # format: {<id>: {'nodes': [...], 'details': {'dvm_uri': <uri>}}}
        # LM specific parameters will be kept under `'details'`
        partitions     = {}

        # go through every dvm instance
        for dvm_id in range(dvm_count):

            node_list = rm.node_list[dvm_id       * nodes_per_dvm:
                                     (dvm_id + 1) * nodes_per_dvm]
            furi      = '%s/prrte.%s.uri'   % (os.getcwd(), dvm_id)
            fhosts    = '%s/prrte.%s.hosts' % (os.getcwd(), dvm_id)
            with open(fhosts, 'w') as fout:
                num_slots = rm.cores_per_node * rm.smt
                for node in node_list:
                    fout.write('%s slots=%d\n' % (node[0], num_slots))
            vm_size = len(node_list)

            prte  = '%s'               % prte_cmd
            prte += ' --prefix %s'     % os.environ['PRRTE_PREFIX']
            prte += ' --report-uri %s' % furi
            prte += ' --hostfile %s'   % fhosts

            if profiler.enabled:
                prte += ' --pmca orte_state_base_verbose 1'  # prte profiling

            # large tasks imply large message sizes, we need to account for that
            # FIXME: need to derive the message size from DVM size - smaller
            #        DVMs will never need large messages, as they can't run
            #        large tasks
            prte += ' --pmca ptl_base_max_msg_size %d' % PTL_BASE_MAX_MSG_SIZE
          # prte += ' --pmca rmaps_base_verbose 5'

            # debug mapper problems for large tasks
            if log.isEnabledFor(logging.DEBUG):
                prte += ' -pmca orte_rmaps_base_verbose 100'

            # we apply two temporary tweaks on Summit which should not be needed
            # in the long run:
            #
            # avoid 64 node limit (ssh connection limit)
            prte += ' --pmca plm_rsh_no_tree_spawn 1'

            # ensure 1 ssh per dvm
            prte += ' --pmca plm_rsh_num_concurrent %d' % vm_size

            # Use (g)stdbuf to disable buffering.  We need this to get the
            # "DVM ready" message to ensure DVM startup completion
            #
            # The command seems to be generally available on our Cray's,
            # if not, we can code some home-cooked pty stuff (TODO)
            stdbuf_cmd = ru.which(['stdbuf', 'gstdbuf'])
            if not stdbuf_cmd:
                raise Exception('(g)stdbuf command not found')
            stdbuf_arg = '-oL'

            # Additional (debug) arguments to prte
            verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))
            if verbose:
                debug_strings = ['--debug-devel',
                                 '--pmca odls_base_verbose 100',
                                 '--pmca rml_base_verbose 100']
            else:
                debug_strings = []

            # Base command = (g)stdbuf <args> + prte + prte-args + debug-args
            cmdline  = '%s %s %s ' % (stdbuf_cmd, stdbuf_arg, prte)
            cmdline += ' '.join(debug_strings)
            cmdline  = cmdline.strip()

            log.info('start prte-%s on %d nodes [%s]', dvm_id, vm_size, cmdline)
            profiler.prof(event='dvm_start',
                          uid=cfg['pid'],
                          msg='dvm_id=%s' % dvm_id)

            dvm_uri     = None
            dvm_process = mp.Popen(cmdline.split(),
                                   stdout=mp.PIPE,
                                   stderr=mp.STDOUT)
            dvm_watcher = mt.Thread(target=cls._watch_dvm,
                                    args=[dvm_process, dvm_id, log])
            dvm_watcher.daemon = True
            dvm_watcher.start()

            for _ in range(100):
                time.sleep(.5)

                try:
                    with open(furi, 'r') as fin:
                        for line in fin.readlines():
                            if '://' in line:
                                dvm_uri = line.strip()
                                break
                except Exception as e:
                    log.debug('DVM check: uri file missing: %s...', str(e)[:24])
                    time.sleep(.5)

                if dvm_uri:
                    break

            if not dvm_uri:
                raise Exception('DVM_URI not found!')

            log.info('prte-%s startup successful: [%s]', dvm_id, dvm_uri)

            profiler.prof(event='dvm_ok',
                          uid=cfg['pid'],
                          msg='dvm_id=%s' % dvm_id)

            partitions[str(dvm_id)] = {
                'nodes'  : [node[1] for node in node_list],
                'details': {'dvm_uri': dvm_uri}}

        lm_info = {'partitions'  : partitions,
                   'version_info': prte_info,
                   'cvd_id_mode' : 'physical'}

        # extra time to allow the DVM(s) to stabilize
        time.sleep(10)

        # we need to inform the actual LaunchMethod instance about the prte URI.
        # So we pass it back to the ResourceManager which will keep it in an
        # 'lm_info', which will then be passed as part of the slots via the
        # scheduler
        return lm_info


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_shutdown_hook(cls, name, cfg, rm, lm_info, log, profiler):
        """
        This hook is symmetric to the config hook above, and is called during
        shutdown sequence, for the sake of freeing allocated resources.
        """
        prun = ru.which('prun')
        if not prun:
            raise Exception("Couldn't find prun")

        for p_id, p_data in lm_info.get('partitions', {}).items():
            dvm_uri = p_data['details']['dvm_uri']
            try:
                log.info('terminating prte-%s (%s)', p_id, dvm_uri)
                ru.sh_callout('%s --hnp %s --terminate' % (prun, dvm_uri))
                profiler.prof(event='dvm_stop',
                              uid=cfg['pid'],
                              msg='dvm_id=%s' % p_id)
            except Exception as e:
                # use the same event name as for runtime failures - those are
                # not distinguishable at the moment from termination failures
                profiler.prof(event='dvm_fail',
                              uid=cfg['pid'],
                              msg='dvm_id=%s | error: %s' % (p_id, e))
                log.debug('prte-%s termination failed (%s)', p_id, dvm_uri)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # ensure that `prun` is in the path (`which` will raise otherwise)
        ru.which('prun')
        self.launch_command = 'prun'


    # --------------------------------------------------------------------------
    #
    def construct_command(self, t, launch_script_hop):

        time.sleep(0.1)

        slots        = t['slots']
        td           = t['description']
        task_exec    = td['executable']
        task_env     = td.get('environment') or dict()
        task_args    = td.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        n_procs      = t['description'].get('cpu_processes') or 1
        n_threads    = t['description'].get('cpu_threads')   or 1

        self._log.debug('prep %s', t['uid'])

        if 'lm_info' not in slots:
            raise RuntimeError('No lm_info to launch via %s: %s'
                               % (self.name, slots))

        if not slots['lm_info']:
            raise RuntimeError('lm_info missing for %s: %s'
                               % (self.name, slots))

        if 'partitions' not in slots['lm_info']:
            raise RuntimeError('no partitions in lm_info for %s: %s'
                               % (self.name, slots))

        if task_argstr: task_command = '%s %s' % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
        env_list   = self.EXPORT_ENV_VARIABLES + list(task_env.keys())
        if env_list:
            for var in env_list:
                env_string += '-x "%s" ' % var

        map_flag  = ' -np %d --cpus-per-proc %d' % (n_procs, n_threads)
        map_flag += ' --bind-to hwthread:overload-allowed --use-hwthread-cpus'
        map_flag += ' --oversubscribe'

        # see DVM startup
        map_flag += ' --pmca ptl_base_max_msg_size %d' % PTL_BASE_MAX_MSG_SIZE
      # map_flag += ' --pmca rmaps_base_verbose 5'

        dvm_uri = ''
        if 'nodes' not in slots:
            # this task is unscheduled - we leave it to PRRTE/PMI-X to
            # correctly place the task
            pass

        else:
            # FIXME: ensure correct binding for procs and threads via slotfile

            # enact the scheduler's host placement.  For now, we leave socket,
            # core and thread placement to the prted, and just add all process
            # slots to the host list.
            hosts = [node['name'] for node in slots['nodes']]
            map_flag += ' -host %s' % ','.join(hosts)
            # another way to provide list of hosts:
            #    map_flag += ' -host %s:%s' % (hosts[0], n_procs)

            # scheduler makes sure that all hosts are from the same DVM
            _host_uid = slots['nodes'][0]['uid']
            for partition in slots['lm_info']['partitions'].values():
                if _host_uid in partition['nodes']:
                    dvm_uri = '--hnp "%s"' % partition['details']['dvm_uri']
                    break

        # Additional (debug) arguments to prun
        if self._verbose:
            debug_string = ' '.join(['--verbose',
                                     # '--debug-devel',
                                     # '-display-devel-map',
                                     # '-display-allocation',
                                     '--report-bindings'])
        else:
            debug_string = '--verbose'  # needed to get prte profile events

      # env_string = ''  # FIXME
        command = '%s %s %s %s %s %s' % (self.launch_command,
                  dvm_uri, map_flag, debug_string, env_string, task_command)

        return command, None


# ------------------------------------------------------------------------------

